# Copyright 2021 The Cirq Developers
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=missing-function-docstring
# pylint: disable=missing-return-doc
# mypy: disable-error-code=method-assign
from __future__ import annotations

import itertools
import pathlib
import re
from unittest.mock import MagicMock

import cirq
import numpy as np
import pandas as pd
import pytest

from supermarq.qcvv import XEB, XEBResults


def test_xeb_init() -> None:
    q0, q1, q2 = cirq.LineQubit.range(3)

    experiment = XEB(num_circuits=10, cycle_depths=[1, 3, 5])
    assert experiment.num_qubits == 2
    assert experiment.qubits == [q0, q1]
    assert experiment.two_qubit_gate == cirq.CZ
    assert experiment.single_qubit_gate_set == [
        cirq.PhasedXZGate(
            z_exponent=z,
            x_exponent=0.5,
            axis_phase_exponent=a,
        )
        for a, z in itertools.product(np.linspace(start=0, stop=7 / 4, num=8), repeat=2)
    ]

    experiment = XEB(two_qubit_gate=cirq.CX, num_circuits=10, cycle_depths=[1, 3, 5])
    assert experiment.num_qubits == 2
    assert experiment.qubits == [q0, q1]
    assert experiment.two_qubit_gate == cirq.CX
    assert experiment.single_qubit_gate_set == [
        cirq.PhasedXZGate(
            z_exponent=z,
            x_exponent=0.5,
            axis_phase_exponent=a,
        )
        for a, z in itertools.product(np.linspace(start=0, stop=7 / 4, num=8), repeat=2)
    ]

    experiment = XEB(single_qubit_gate_set=[cirq.X], num_circuits=10, cycle_depths=[1, 3, 5])
    assert experiment.num_qubits == 2
    assert experiment.two_qubit_gate == cirq.CZ
    assert experiment.single_qubit_gate_set == [cirq.X]

    experiment = XEB(two_qubit_gate=None, num_circuits=10, cycle_depths=[1, 3, 5])
    assert experiment.num_qubits == 2
    assert experiment.qubits == [q0, q1]
    assert experiment.two_qubit_gate is None

    experiment = XEB(two_qubit_gate=cirq.CZ(q1, q2), num_circuits=10, cycle_depths=[1, 3, 5])
    assert experiment.num_qubits == 2
    assert experiment.qubits == [q1, q2]
    assert experiment.two_qubit_gate == cirq.CZ
    assert all(sample.circuit.all_qubits() == {q1, q2} for sample in experiment.samples)


@pytest.fixture
def xeb_experiment() -> XEB:
    return XEB(
        single_qubit_gate_set=[cirq.X, cirq.Y, cirq.Z], num_circuits=10, cycle_depths=[1, 3, 5]
    )


def test_build_xeb_circuit(xeb_experiment: XEB) -> None:

    xeb_experiment._rng = (rng := MagicMock())
    rng.choice.side_effect = [
        np.array([[cirq.X, cirq.Y], [cirq.Z, cirq.Y], [cirq.Y, cirq.Z]]),
        np.array([[cirq.X, cirq.Z], [cirq.X, cirq.X], [cirq.Y, cirq.Y]]),
    ]
    circuits = xeb_experiment._build_circuits(num_circuits=2, cycle_depths=[2])

    assert len(circuits) == 2

    qbs = xeb_experiment.qubits
    cirq.testing.assert_same_circuits(
        circuits[0].circuit,
        cirq.Circuit(
            [
                cirq.X(qbs[0]),
                cirq.Y(qbs[1]),
                cirq.TaggedOperation(cirq.CZ(*qbs), "no_compile"),
                cirq.Z(qbs[0]),
                cirq.Y(qbs[1]),
                cirq.TaggedOperation(cirq.CZ(*qbs), "no_compile"),
                cirq.Y(qbs[0]),
                cirq.Z(qbs[1]),
                cirq.measure(qbs),
            ]
        ),
    )
    assert circuits[0].data == {
        "circuit_depth": 5,
        "cycle_depth": 2,
        "two_qubit_gate": "CZ",
        "exact_00": 1.0,
        "exact_01": 0.0,
        "exact_10": 0.0,
        "exact_11": 0.0,
    }
    cirq.testing.assert_same_circuits(
        circuits[1].circuit,
        cirq.Circuit(
            [
                cirq.X(qbs[0]),
                cirq.Z(qbs[1]),
                cirq.TaggedOperation(cirq.CZ(*qbs), "no_compile"),
                cirq.X(qbs[0]),
                cirq.X(qbs[1]),
                cirq.TaggedOperation(cirq.CZ(*qbs), "no_compile"),
                cirq.Y(qbs[0]),
                cirq.Y(qbs[1]),
                cirq.measure(qbs),
            ]
        ),
    )
    assert circuits[1].data == {
        "circuit_depth": 5,
        "cycle_depth": 2,
        "two_qubit_gate": "CZ",
        "exact_00": 0.0,
        "exact_01": 0.0,
        "exact_10": 1.0,
        "exact_11": 0.0,
    }


def test_xeb_analyse_results(tmp_path: pathlib.Path, xeb_experiment: XEB) -> None:
    results = XEBResults(target="example", experiment=xeb_experiment)

    results.data = pd.DataFrame(
        [
            {
                "circuit_realization": 0,
                "cycle_depth": 1,
                "circuit_depth": 3,
                "00": 1.0,
                "01": 0.0,
                "10": 0.0,
                "11": 0.0,
                "exact_00": 1.0,
                "exact_01": 0.0,
                "exact_10": 0.0,
                "exact_11": 0.0,
            },
            {
                "circuit_realization": 1,
                "cycle_depth": 1,
                "circuit_depth": 3,
                "00": 1.0,
                "01": 0.0,
                "10": 0.0,
                "11": 0.0,
                "exact_00": 0.5,
                "exact_01": 0.5,
                "exact_10": 0.0,
                "exact_11": 0.0,
            },
            {
                "circuit_realization": 0,
                "cycle_depth": 5,
                "circuit_depth": 11,
                "00": 0.0,
                "01": 1.0,
                "10": 0.0,
                "11": 0.0,
                "exact_00": 0.0,
                "exact_01": 0.75,
                "exact_10": 0.25,
                "exact_11": 0.0,
            },
            {
                "circuit_realization": 1,
                "cycle_depth": 5,
                "circuit_depth": 11,
                "00": 0.0,
                "01": 0.0,
                "10": 1.0,
                "11": 0.0,
                "exact_00": 0.0,
                "exact_01": 0.5,
                "exact_10": 0.25,
                "exact_11": 0.25,
            },
            {
                "circuit_realization": 0,
                "cycle_depth": 10,
                "circuit_depth": 21,
                "00": 0.0,
                "01": 0.0,
                "10": 0.5,
                "11": 0.5,
                "exact_00": 0.2,
                "exact_01": 0.3,
                "exact_10": 0.25,
                "exact_11": 0.25,
            },
            {
                "circuit_realization": 1,
                "cycle_depth": 10,
                "circuit_depth": 21,
                "00": 0.0,
                "01": 0.0,
                "10": 0.5,
                "11": 0.5,
                "exact_00": 0.1,
                "exact_01": 0.1,
                "exact_10": 0.4,
                "exact_11": 0.4,
            },
        ]
    )

    plot_filename = tmp_path / "example.png"
    speckle_plot_filename = tmp_path / "example_speckle.png"

    results.analyze(plot_filename=plot_filename.as_posix())
    np.testing.assert_allclose(
        results.data["sum_p(x)p^(x)"].values, [1.0, 0.5, 0.75, 0.25, 0.25, 0.4]
    )
    np.testing.assert_allclose(
        results.data["sum_p(x)p(x)"].values, [1.0, 0.5, 0.625, 0.375, 0.255, 0.34]
    )

    # Calculated by hand
    assert results.cycle_fidelity_estimate == pytest.approx(1.0613025)
    assert results.cycle_fidelity_estimate_std == pytest.approx(0.0597633930)

    assert pathlib.Path(tmp_path / "example.png").exists()

    # Test the speckle plot
    results.plot_speckle(filename=speckle_plot_filename.as_posix())
    assert pathlib.Path(tmp_path / "example_speckle.png").exists()


def test_results_no_data() -> None:
    results = XEBResults(target="example", experiment=XEB(1, []), data=None)
    with pytest.raises(RuntimeError, match="No data stored. Cannot perform analysis."):
        results._analyze()

    with pytest.raises(RuntimeError, match="No data stored. Cannot plot results."):
        results.plot_results()

    with pytest.raises(RuntimeError, match="No data stored. Cannot plot results."):
        results.plot_speckle()

    with pytest.raises(
        RuntimeError, match="No stored dataframe of circuit fidelities. Something has gone wrong."
    ):
        results.data = pd.DataFrame()
        results.plot_results()


def test_results_not_analyzed() -> None:
    results = XEBResults(target="example", experiment=XEB(1, []), data=None)
    for attr in ["cycle_fidelity_estimate", "cycle_fidelity_estimate_std"]:
        with pytest.raises(
            RuntimeError,
            match=re.escape("Value has not yet been estimated. Please run `.analyze()` method."),
        ):
            getattr(results, attr)


def test_dump_and_load(
    tmp_path_factory: pytest.TempPathFactory,
    xeb_experiment: XEB,
) -> None:
    filename = tmp_path_factory.mktemp("tempdir") / "file.json"
    xeb_experiment.to_file(filename)
    exp = XEB.from_file(filename)

    assert exp.samples == xeb_experiment.samples
    assert exp.num_qubits == xeb_experiment.num_qubits
    assert exp.num_circuits == xeb_experiment.num_circuits
    assert exp.cycle_depths == xeb_experiment.cycle_depths
    assert exp.single_qubit_gate_set == xeb_experiment.single_qubit_gate_set
    assert exp.two_qubit_gate == xeb_experiment.two_qubit_gate



# pylint: disable=missing-function-docstring,missing-class-docstring
from __future__ import annotations

import cirq

import supermarq

CIRCUIT = cirq.Circuit(
    cirq.SWAP(*cirq.LineQubit.range(2)),
    cirq.measure(cirq.LineQubit(0)),
    cirq.reset(cirq.LineQubit(0)),
    cirq.measure(*cirq.LineQubit.range(2)),
)


def test_compute_communication() -> None:
    feature = supermarq.features.compute_communication(CIRCUIT)
    assert feature >= 0 and feature <= 1


def test_compute_liveness() -> None:
    feature = supermarq.features.compute_liveness(CIRCUIT)
    assert feature >= 0 and feature <= 1


def test_compute_parallelism() -> None:
    feature = supermarq.features.compute_parallelism(CIRCUIT)
    assert feature >= 0 and feature <= 1

    assert supermarq.features.compute_parallelism(cirq.Circuit()) == 0


def test_compute_measurement() -> None:
    feature = supermarq.features.compute_measurement(CIRCUIT)
    assert feature >= 0 and feature <= 1


def test_compute_entanglement() -> None:
    feature = supermarq.features.compute_entanglement(CIRCUIT)
    assert feature >= 0 and feature <= 1


def test_compute_depth() -> None:
    qubits = cirq.LineQubit.range(4)
    test_circuit = cirq.Circuit(
        cirq.CX(qubits[0], qubits[1]),
        cirq.CZ(qubits[2], qubits[3]),
        cirq.CX(qubits[1], qubits[2]),
        cirq.CX(qubits[2], qubits[3]),
    )
    test_feature = supermarq.features.compute_depth(test_circuit)
    assert test_feature >= 0 and test_feature <= 1

    assert supermarq.features.compute_depth(cirq.Circuit()) == 0



# Copyright 2021 The Cirq Developers
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tooling for interleaved randomised benchmarking"""

from __future__ import annotations

from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import cirq
import cirq.circuits
import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import scipy
import seaborn as sns
from tqdm.auto import trange
from tqdm.contrib.itertools import product

from supermarq.qcvv.base_experiment import QCVVExperiment, QCVVResults, Sample

if TYPE_CHECKING:
    from typing_extensions import Self


####################################################################################################
# Some handy functions for 1 and 2 qubit Clifford operations
####################################################################################################
def _reduce_single_qubit_clifford_seq(
    gate_seq: list[cirq.SingleQubitCliffordGate],
) -> cirq.SingleQubitCliffordGate:
    """Reduces a list of single qubit clifford gates to a single gate.

    Args:
        gate_seq: The list of gates.
    Returns:
        The single reduced gate.
    """
    cur = gate_seq[0]
    for gate in gate_seq[1:]:
        cur = cur.merged_with(gate)
    return cur


def _reduce_clifford_seq(
    gate_seq: list[cirq.CliffordGate],
) -> cirq.CliffordGate:
    """Reduces a list of multi qubit clifford gates to a single gate.

    Args:
        gate_seq: The list of gates.
    Returns:
        The single reduced gate.
    """
    cur = gate_seq[0].clifford_tableau
    for gate in gate_seq[1:]:
        cur = cur.then(gate.clifford_tableau)
    return cirq.CliffordGate.from_clifford_tableau(cur)


####################################################################################################
# The sets `S1`, `S1_X` and `S1_Y` of single qubit Clifford operations are used to generate
# random two qubit Clifford operations. For details see: https://arxiv.org/abs/1210.7011 &
# https://arxiv.org/abs/1402.4848.
# The implementation is adapted from:
# https://github.com/quantumlib/Cirq/blob/main/cirq-core/cirq/experiments/qubit_characterizations.py
####################################################################################################
_S1 = [
    # I
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([False, False], dtype=np.dtype("bool")),
            xs=np.array([[True], [False]], dtype=np.dtype("bool")),
            zs=np.array([[False], [True]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
    # Y**0.5 X**0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([False, False], dtype=np.dtype("bool")),
            xs=np.array([[True], [True]], dtype=np.dtype("bool")),
            zs=np.array([[True], [False]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
    # X**-0.5 Y**-0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([False, False], dtype=np.dtype("bool")),
            xs=np.array([[False], [True]], dtype=np.dtype("bool")),
            zs=np.array([[True], [True]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
]


_S1_X = [
    # X**0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([False, True], dtype=np.dtype("bool")),
            xs=np.array([[True], [True]], dtype=np.dtype("bool")),
            zs=np.array([[False], [True]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
    # X**0.5, Y**0.5, X**0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([False, True], dtype=np.dtype("bool")),
            xs=np.array([[True], [False]], dtype=np.dtype("bool")),
            zs=np.array([[True], [True]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
    # Y**-0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([False, True], dtype=np.dtype("bool")),
            xs=np.array([[False], [True]], dtype=np.dtype("bool")),
            zs=np.array([[True], [False]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
]


_S1_Y = [
    # Y**0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([True, False], dtype=np.dtype("bool")),
            xs=np.array([[False], [True]], dtype=np.dtype("bool")),
            zs=np.array([[True], [False]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
    # X**-0.5 Y**-0.5 X**0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([True, False], dtype=np.dtype("bool")),
            xs=np.array([[True], [False]], dtype=np.dtype("bool")),
            zs=np.array([[True], [True]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
    # Y X**0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([True, False], dtype=np.dtype("bool")),
            xs=np.array([[True], [True]], dtype=np.dtype("bool")),
            zs=np.array([[False], [True]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
]


@dataclass
class _RBResultsBase(QCVVResults):
    _rb_decay_coefficient: float | None = None
    """Decay coefficient estimate without the interleaving gate."""
    _rb_decay_coefficient_std: float | None = None
    """Standard deviation of the decay coefficient estimate without the interleaving gate."""

    @property
    def rb_decay_coefficient(self) -> float:
        """Returns:
        Decay coefficient estimate without the interleaving gate."""
        if self._rb_decay_coefficient is None:
            raise self._not_analyzed
        return self._rb_decay_coefficient

    @property
    def rb_decay_coefficient_std(self) -> float:
        """Returns:
        Standard deviation of the decay coefficient estimate without the interleaving gate."""
        if self._rb_decay_coefficient_std is None:
            raise self._not_analyzed
        return self._rb_decay_coefficient_std

    def _fit_decay(
        self, experiment: str = "RB"
    ) -> tuple[np.typing.NDArray[np.float64], np.typing.NDArray[np.float64]]:
        """Fits the exponential decay function to either the RB or IRB results.

        Args:
            experiment: Either `RB` or `IRB` referring to which data to filter by. Defaults to "RB".

        Returns:
            Tuple of the decay fit parameters and their standard deviations.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot perform fit.")

        xx = self.data.query(f"experiment == '{experiment}'")["clifford_depth"]
        yy = self.data.query(f"experiment == '{experiment}'")["0" * self.num_qubits]

        popt, pcov = scipy.optimize.curve_fit(
            self.exp_decay,
            xx,
            yy,
            p0=(1.0 - 2**-self.num_qubits, 0.99, 2**-self.num_qubits),
            bounds=(0, 1),
            max_nfev=2000,
        )

        return popt, np.sqrt(np.diag(pcov))

    @staticmethod
    def exp_decay(x: npt.ArrayLike, A: float, alpha: float, B: float) -> npt.ArrayLike:
        r"""Exponential decay of the form

        .. math::

            A \alpha^x + B

        Args:
            x: x
            A: Decay constant
            alpha: Decay coefficient
            B: Additive constant

        Returns:
            Exponential decay applied to x.
        """
        return A * (np.asarray(alpha) ** x) + B

    def _plot_results(self) -> plt.Axes:
        """Plot the exponential decay of the circuit fidelity with cycle depth.

        Returns:
            A matplotlib axiss containing the RB decay plots and the corresponding
            fit.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot make plot.")
        plot = sns.scatterplot(
            data=self.data,
            x="clifford_depth",
            y="0" * self.num_qubits,
            hue="experiment",
        )
        plot.set_xlabel(r"Cycle depth", fontsize=15)
        plot.set_ylabel(r"Survival probability", fontsize=15)
        plot.set_title(r"Exponential decay of survival probability", fontsize=15)

        rb_fit = self._fit_decay()
        xx = np.linspace(0, np.max(self.data.clifford_depth))
        plot.plot(
            xx,
            self.exp_decay(xx, *rb_fit[0]),
            color="tab:blue",
            linestyle="--",
        )
        plot.fill_between(
            xx,
            self.exp_decay(xx, *(rb_fit[0] - rb_fit[1])),
            self.exp_decay(xx, *(rb_fit[0] + rb_fit[1])),
            alpha=0.5,
            color="tab:blue",
        )

        return plot

    def _analyze(self) -> None:
        rb_fit = self._fit_decay("RB")
        self._rb_decay_coefficient, self._rb_decay_coefficient_std = rb_fit[0][1], rb_fit[1][1]


@dataclass
class IRBResults(_RBResultsBase):
    """Data structure for the IRB experiment results."""

    _irb_decay_coefficient: float | None = None
    """Decay coefficient estimate with the interleaving gate."""
    _irb_decay_coefficient_std: float | None = None
    """Standard deviation of the decay coefficient estimate with the interleaving gate."""
    _average_interleaved_gate_error: float | None = None
    """Estimate of the interleaving gate error."""
    _average_interleaved_gate_error_std: float | None = None
    """Standard deviation of the estimate for the interleaving gate error."""

    @property
    def irb_decay_coefficient(self) -> float:
        """Returns:
        Decay coefficient estimate with the interleaving gate."""
        if self._irb_decay_coefficient is None:
            raise self._not_analyzed
        return self._irb_decay_coefficient

    @property
    def irb_decay_coefficient_std(self) -> float:
        """Returns:
        Standard deviation of the decay coefficient estimate with the interleaving gate."""
        if self._irb_decay_coefficient_std is None:
            raise self._not_analyzed
        return self._irb_decay_coefficient_std

    @property
    def average_interleaved_gate_error(self) -> float:
        """Returns:
        Estimate of the interleaving gate error."""
        if self._average_interleaved_gate_error is None:
            raise self._not_analyzed
        return self._average_interleaved_gate_error

    @property
    def average_interleaved_gate_error_std(self) -> float:
        """Returns:
        Standard deviation of the estimate for the interleaving gate error."""
        if self._average_interleaved_gate_error_std is None:
            raise self._not_analyzed
        return self._average_interleaved_gate_error_std

    def plot_results(
        self,
        filename: str | None = None,
    ) -> plt.Figure:
        """Plot the exponential decay of the circuit fidelity with cycle depth.

        Args:
            filename: Optional argument providing a filename to save the plots to. Defaults to None,
                indicating not to save the plot.

        Returns:
            A single matplotlib figure containing the IRB and RB decay plots and the corresponding
            fits.

        Raises:
            RuntimeError: If no data is stored.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot make plot.")
        plot = self._plot_results()
        irb_fit = self._fit_decay("IRB")
        xx = np.linspace(0, np.max(self.data.clifford_depth))
        plot.plot(
            xx,
            self.exp_decay(xx, *irb_fit[0]),
            color="tab:orange",
            linestyle="--",
        )
        plot.fill_between(
            xx,
            self.exp_decay(xx, *(irb_fit[0] - irb_fit[1])),
            self.exp_decay(xx, *(irb_fit[0] + irb_fit[1])),
            alpha=0.5,
            color="tab:orange",
        )
        if filename is not None:
            plt.savefig(filename)

        root_figure = plot.get_figure()
        assert isinstance(root_figure, plt.Figure)
        if filename is not None:
            root_figure.savefig(filename, bbox_inches="tight")
        return root_figure

    def _analyze(self) -> None:
        super()._analyze()

        irb_fit = self._fit_decay("IRB")
        irb_decay_coefficient, irb_decay_coefficient_std = irb_fit[0][1], irb_fit[1][1]
        interleaved_gate_error = (1 - irb_decay_coefficient / self.rb_decay_coefficient) * (
            1 - 2**-self.num_qubits
        )

        interleaved_gate_error_std = (
            (1 - 2**-self.num_qubits) / self.rb_decay_coefficient
        ) * np.sqrt(
            irb_decay_coefficient_std**2
            + irb_decay_coefficient**2 * self.rb_decay_coefficient_std**2
        )

        self._irb_decay_coefficient = irb_decay_coefficient
        self._irb_decay_coefficient_std = irb_decay_coefficient_std
        self._average_interleaved_gate_error = interleaved_gate_error
        self._average_interleaved_gate_error_std = interleaved_gate_error_std

    def print_results(self) -> None:
        print(
            f"Estimated gate error: {self.average_interleaved_gate_error:.6f} +/- "
            f"{self.average_interleaved_gate_error_std:.6f}"
        )


@dataclass
class RBResults(_RBResultsBase):
    """Data structure for the RB experiment results."""

    _average_error_per_clifford: float | None = None
    """Estimate of the average error per Clifford operation."""
    _average_error_per_clifford_std: float | None = None
    """Standard deviation of the the average error per Clifford operation."""

    @property
    def average_error_per_clifford(self) -> float:
        """Returns:
        Estimate of the average error per Clifford operation."""
        if self._average_error_per_clifford is None:
            raise self._not_analyzed
        return self._average_error_per_clifford

    @property
    def average_error_per_clifford_std(self) -> float:
        """Returns:
        Standard deviation of the the average error per Clifford operation."""
        if self._average_error_per_clifford_std is None:
            raise self._not_analyzed
        return self._average_error_per_clifford_std

    def plot_results(
        self,
        filename: str | None = None,
    ) -> plt.Figure:
        """Plot the exponential decay of the circuit fidelity with cycle depth.

        Args:
            filename: Optional argument providing a filename to save the plots to. Defaults to None,
                indicating not to save the plot.

        Returns:
            A single matplotlib figure containing the RB decay plot and the corresponding fit.

        Raises:
            RuntimeError: If no data is stored.
        """
        if self.data is None:
            raise RuntimeError("No data stored. Cannot make plot.")

        plot = self._plot_results()
        root_figure = plot.get_figure()
        assert isinstance(root_figure, plt.Figure)
        if filename is not None:
            root_figure.savefig(filename, bbox_inches="tight")
        return root_figure

    def _analyze(self) -> None:
        super()._analyze()
        self._average_error_per_clifford = (1 - 2**-self.num_qubits) * (
            1 - self.rb_decay_coefficient
        )
        self._average_error_per_clifford_std = (
            1 - 2**-self.num_qubits
        ) * self.rb_decay_coefficient_std

    def print_results(self) -> None:

        print(
            f"Estimated error per Clifford: {self.average_error_per_clifford:.6f} +/- "
            f"{self.average_error_per_clifford_std:.6f}"
        )


class IRB(QCVVExperiment[_RBResultsBase]):
    r"""Interleaved random benchmarking (IRB) experiment.

    IRB estimates the gate error of specified Clifford gate, :math:`\mathcal{C}^*`.
    This is achieved by first choosing a random sequence, :math:`\{\mathcal{C_i}\}_m`
    of :math:`m` Clifford gates and then using this to generate two circuits. The first
    is generated by appending to this sequence the single gate that corresponds to the
    inverse of the original sequence. The second circuit it obtained by inserting the
    interleaving gate, :math:`\mathcal{C}^*` after each gate in the sequence and then
    again appending the corresponding inverse element of the new circuit. Thus both
    circuits correspond to the identity operation.

    We run both circuits on the specified target and calculate the probability of measuring
    the resulting state in the ground state, :math:`p(0...0)`. This gives the circuit fidelity

    .. math::

        f(m) = 2p(0...0) - 1

    We can then fit an exponential decay :math:`\log(f) \sim m` to this circuit fidelity
    for each circuit, with decay rates :math:`\alpha` and :math:`\tilde{\alpha}` for the circuit
    without and with interleaving respectively. Finally the gate error of the
    specified gate, :math:`\mathcal{C}^*` is estimated as

    .. math::

        e_{\mathcal{C}^*} = \frac{1}{2} \left(1 - \frac{\tilde{\alpha}}{\alpha}\right)

    For more details see: https://arxiv.org/abs/1203.4550
    """

    def __init__(
        self,
        num_circuits: int,
        cycle_depths: Iterable[int],
        interleaved_gate: cirq.Gate | None = cirq.Z,
        qubits: int | Sequence[cirq.Qid] = 1,
        clifford_op_gateset: cirq.CompilationTargetGateset = cirq.CZTargetGateset(),
        *,
        random_seed: int | np.random.Generator | None = None,
        _samples: list[Sample] | None = None,
        **kwargs: str,
    ) -> None:
        """Constructs an IRB experiment.

        Args:
            num_circuits: Number of circuits to sample.
            cycle_depths: The cycle depths to sample.
            interleaved_gate: The Clifford gate to measure the gate error of. If None then no
                interleaving is performed and instead vanilla randomized benchmarking is performed.
            qubits: The qubit(s) to experiment on. If an integer, must either be 1 or 2; otherwise
                must be a sequence of 1 or 2 qubit(s). Ignored if a gate is provided - the qubits
                are instead inferred from the gate.
            clifford_op_gateset: The gateset to use when implementing the clifford operations.
                Defaults to the CZ/GR set.
            random_seed: An optional seed to use for randomization.
        """
        if isinstance(interleaved_gate, cirq.Operation):
            qubits = interleaved_gate.qubits
            interleaved_gate = interleaved_gate.gate
        elif interleaved_gate is not None:
            qubits = cirq.LineQubit.range(cirq.num_qubits(interleaved_gate))
        elif not isinstance(qubits, Sequence):
            qubits = cirq.LineQubit.range(qubits)

        if len(qubits) not in [1, 2]:
            raise NotImplementedError(
                "IRB experiment is currently only implemented for single or two qubit use."
            )

        if interleaved_gate is not None:
            self.interleaved_gate: cirq.CliffordGate | None = cirq.CliffordGate.from_op_list(
                [interleaved_gate(*qubits)], qubits
            )
        else:
            self.interleaved_gate = None

        self.clifford_op_gateset = clifford_op_gateset
        """The gateset to use when implementing Clifford operations."""

        if self.interleaved_gate is None:
            results_cls: type[RBResults] | type[IRBResults] = RBResults
        else:
            results_cls = IRBResults

        super().__init__(
            qubits=qubits,
            num_circuits=num_circuits,
            cycle_depths=cycle_depths,
            random_seed=random_seed,
            results_cls=results_cls,
            _samples=_samples,
            **kwargs,
        )

    def _clifford_gate_to_circuit(
        self,
        clifford: cirq.CliffordGate,
    ) -> cirq.Circuit:
        """Converts a Clifford gate to a circuit using the desired gateset for the experiment.

        Args:
            clifford: The clifford operation to convert.

        Returns:
            A circuit implementing the desired Clifford gate.
        """
        circuit = cirq.Circuit(
            cirq.decompose_clifford_tableau_to_operations(
                list(self.qubits), clifford.clifford_tableau
            )
        )
        return cirq.optimize_for_target_gateset(circuit, gateset=self.clifford_op_gateset)

    def random_single_qubit_clifford(self) -> cirq.SingleQubitCliffordGate:
        """Choose a random single qubit clifford gate.

        Returns:
            The random clifford gate.
        """
        Id = cirq.SingleQubitCliffordGate.I
        H = cirq.SingleQubitCliffordGate.H
        S = cirq.SingleQubitCliffordGate.Z_sqrt
        X = cirq.SingleQubitCliffordGate.X
        Y = cirq.SingleQubitCliffordGate.Y
        Z = cirq.SingleQubitCliffordGate.Z

        set_A = np.array(
            [
                Id,
                S,
                H,
                _reduce_single_qubit_clifford_seq([H, S]),
                _reduce_single_qubit_clifford_seq([S, H]),
                _reduce_single_qubit_clifford_seq([H, S, H]),
            ]
        )

        set_B = np.array([Id, X, Y, Z])

        return _reduce_single_qubit_clifford_seq([self._rng.choice(set_A), self._rng.choice(set_B)])

    def random_two_qubit_clifford(self) -> cirq.CliffordGate:
        """Choose a random two qubit clifford gate.

        For algorithm details see https://arxiv.org/abs/1402.4848 & https://arxiv.org/abs/1210.7011.

        Returns:
            The random clifford gate.
        """
        qubits = cirq.LineQubit.range(2)
        a = self.random_single_qubit_clifford()
        b = self.random_single_qubit_clifford()
        idx = self._rng.integers(20)
        if idx == 0:
            return cirq.CliffordGate.from_op_list([a(qubits[0]), b(qubits[1])], qubits)
        elif idx == 1:
            return cirq.CliffordGate.from_op_list(
                [
                    a(qubits[0]),
                    b(qubits[1]),
                    cirq.CZ(*qubits),
                    cirq.Y(qubits[0]) ** -0.5,
                    cirq.Y(qubits[1]) ** 0.5,
                    cirq.CZ(*qubits),
                    cirq.Y(qubits[0]) ** 0.5,
                    cirq.Y(qubits[1]) ** -0.5,
                    cirq.CZ(*qubits),
                    cirq.Y(qubits[1]) ** 0.5,
                ],
                qubits,
            )
        elif 2 <= idx <= 10:
            idx_a = int((idx - 2) / 3)
            idx_b = (idx - 2) % 3
            return cirq.CliffordGate.from_op_list(
                [
                    a(qubits[0]),
                    b(qubits[1]),
                    cirq.CZ(*qubits),
                    _S1[idx_a](qubits[0]),
                    _S1_Y[idx_b](qubits[1]),
                ],
                qubits,
            )

        idx_a = int((idx - 11) / 3)
        idx_b = (idx - 11) % 3
        return cirq.CliffordGate.from_op_list(
            [
                a(qubits[0]),
                b(qubits[1]),
                cirq.CZ(*qubits),
                cirq.Y(qubits[0]) ** 0.5,
                cirq.X(qubits[1]) ** -0.5,
                cirq.CZ(*qubits),
                _S1_Y[idx_a](qubits[0]),
                _S1_X[idx_b](qubits[1]),
            ],
            qubits,
        )

    def random_clifford(self) -> cirq.CliffordGate:
        """Returns:
        A random clifford gate with the correct number of qubits for the current experiment.
        """
        if self.num_qubits == 1:
            return self.random_single_qubit_clifford()
        return self.random_two_qubit_clifford()

    def gates_per_clifford(self, samples: int = 500) -> dict[str, float]:
        """Samples a number of random Clifford operations and calculates the average number of
        single and two qubit gates used to implement them. Note this depends on the gateset chosen
        for the experiment.

        Args:
            samples: Number of samples to use. Defaults to 500.

        Returns:
            A dictionary with the average number of one and two qubit gates used.
        """
        sample = [
            self._clifford_gate_to_circuit(self.random_clifford())
            for _ in trange(samples, desc="Sampling Clifford operations")
        ]
        return {
            "single_qubit_gates": np.mean(
                [
                    sum(1 for op in circuit.all_operations() if len(op.qubits) == 1)
                    for circuit in sample
                ]
            ).item(),
            "two_qubit_gates": np.mean(
                [
                    sum(1 for op in circuit.all_operations() if len(op.qubits) == 2)
                    for circuit in sample
                ]
            ).item(),
        }

    def _build_circuits(self, num_circuits: int, cycle_depths: Iterable[int]) -> Sequence[Sample]:
        """Build a list of randomised circuits required for the IRB experiment.

        Args:
            num_circuits: Number of circuits to generate.
            cycle_depths: An iterable of the different cycle depths to use during the experiment.

        Returns:
            The list of experiment samples.
        """
        samples = []
        for k, depth in product(range(num_circuits), cycle_depths, desc="Building circuits"):
            base_sequence = [self.random_clifford() for _ in range(depth)]
            rb_sequence = base_sequence + [
                _reduce_clifford_seq(cirq.inverse(base_sequence))  # type: ignore[arg-type]
            ]
            rb_circuit = cirq.Circuit(self._clifford_gate_to_circuit(gate) for gate in rb_sequence)
            samples.append(
                Sample(
                    circuit=rb_circuit + cirq.measure(sorted(self.qubits)),
                    data={
                        "clifford_depth": depth,
                        "circuit_depth": len(rb_circuit),
                        "single_qubit_gates": sum(
                            1 for op in rb_circuit.all_operations() if len(op.qubits) == 1
                        ),
                        "two_qubit_gates": sum(
                            1 for op in rb_circuit.all_operations() if len(op.qubits) == 2
                        ),
                        "experiment": "RB",
                    },
                    circuit_realization=k,
                ),
            )
            if self.interleaved_gate is not None:
                # Find final gate
                irb_sequence = [elem for x in base_sequence for elem in (x, self.interleaved_gate)]
                irb_sequence_final_gate = _reduce_clifford_seq(
                    cirq.inverse(irb_sequence)  # type: ignore[arg-type]
                )

                irb_circuit = cirq.Circuit()
                for gate in base_sequence:
                    irb_circuit += self._clifford_gate_to_circuit(gate)
                    irb_circuit += self.interleaved_gate(*self.qubits).with_tags("no_compile")
                # Add the final inverting gate
                irb_circuit += self._clifford_gate_to_circuit(
                    irb_sequence_final_gate,
                )

                samples.append(
                    Sample(
                        circuit=irb_circuit + cirq.measure(sorted(self.qubits)),
                        data={
                            "clifford_depth": depth,
                            "circuit_depth": len(irb_circuit),
                            "single_qubit_gates": sum(
                                1 for op in irb_circuit.all_operations() if len(op.qubits) == 1
                            ),
                            "two_qubit_gates": sum(
                                1 for op in irb_circuit.all_operations() if len(op.qubits) == 2
                            ),
                            "experiment": "IRB",
                        },
                        circuit_realization=k,
                    ),
                )
        return samples

    def _json_dict_(self) -> dict[str, Any]:
        """Converts the experiment to a json-able dictionary that can be used to recreate the
        experiment object. Note that the state of the random number generator is not stored.

        Returns:
            Json-able dictionary of the experiment data.
        """
        return {
            "interleaved_gate": self.interleaved_gate,
            "clifford_op_gateset": self.clifford_op_gateset,
            **super()._json_dict_(),
        }

    @classmethod
    def _from_json_dict_(
        cls,
        samples: list[Sample],
        interleaved_gate: cirq.Gate,
        clifford_op_gateset: cirq.CompilationTargetGateset,
        num_circuits: int,
        cycle_depths: list[int],
        **kwargs: Any,
    ) -> Self:
        """Creates a experiment from a dictionary of the data.

        Args:
            dictionary: Dict containing the experiment data.

        Returns:
            The deserialized experiment object.
        """
        return cls(
            num_circuits=num_circuits,
            cycle_depths=cycle_depths,
            clifford_op_gateset=clifford_op_gateset,
            interleaved_gate=interleaved_gate,
            _samples=samples,
            **kwargs,
        )



from __future__ import annotations

import copy

import cirq
import numpy as np
import numpy.typing as npt


class MeasurementCircuit:
    """A circuit to simultaneously measure a set of stabilizers.

    The getter and setter functions defined here are used by the Mermin Bell benchmark which
    constructs a simultaneous measurement circuit using a Gaussian elimination process.
    """

    def __init__(
        self,
        circuit: cirq.Circuit,
        stabilizer_matrix: npt.NDArray[np.uint8],
        num_qubits: int,
        qubits: list[cirq.LineQubit],
    ) -> None:
        """Intializes a `MeasurementCircuit`.

        Args:
            circuit: The circuit to measure stabilizers.
            stabilizer_matrix: The stabilizer matrix for the circuit.
            num_qubits: Number of qubits for the circuit.
            qubits: A list of `cirq.LineQubit` instances.
        """
        self.circuit = circuit
        self.stabilizer_matrix = stabilizer_matrix
        self.num_qubits = num_qubits
        self.qubits = qubits

    def get_circuit(self) -> cirq.Circuit:
        """Gets the current circuit.

        Returns:
            The current quantum circuit.
        """
        return self.circuit

    def get_stabilizer(self) -> npt.NDArray[np.uint8]:
        """Gets the current stabilizer matrix.

        The stabilizer matrix is in the Z+X format where M Pauli strings, acting on N qubits, is
        represented as a (2*N, M) matrix.

        For instance, YYI, XXY, IYZ would be represented by
        .. code::

            [[1, 0, 0],  ========
             [1, 0, 1],  Z matrix (top half)
             [0, 1, 1],  ========
             [1, 1, 0],  ========
             [1, 1, 1],  X matrix (bottom half)
             [0, 1, 0]   ========

        Returns:
            The current stabilizer matrix.
        """
        return self.stabilizer_matrix

    def set_circuit(self, circuit: cirq.Circuit) -> None:
        """Assign the class circuit to the input circuit.

        Args:
            circuit: The new circuit which will override the current one.
        """
        self.circuit = circuit

    def set_stabilizer(self, stabilizer_matrix: npt.NDArray[np.uint8]) -> None:
        """The input matrix is assigned to the class matrix.

        Args:
            stabilizer_matrix: An input matrix in X+Z format.
        """
        self.stabilizer_matrix = stabilizer_matrix


def construct_stabilizer(
    N: int, clique: list[tuple[float, str]]
) -> tuple[npt.NDArray[np.uint8], list[str]]:
    """Construct the independent Z+X stabilizer matrix for the given clique.

    All of the terms in the input clique can be measured simultaneously. To construct
    the circuit which will perform this measurement we need to find an independent
    basis for these terms. We'll do this by constructing the Z+X stabilizer matrix
    using all of the terms then select an independent basis using binary Gaussian
    elimination.

    This implementation follows the design of Algorithm 2 in [Minimizing State Preparations
    in Variational Quantum Eigensolver by Partitioning into Commuting
    Families](https://arxiv.org/abs/1907.13623).

    Note:
        This implementation is tailored to the Mermin operator, and assumes that no Pauli Z
        matrices appear in the clique terms. This function will fail if applied to general Pauli
        strings.

    Args:
        N: An integer corresponding to the number of qubits.
        clique: A list of (coefficient, Pauli string) pairs, for example:
            [(-1.2, XXY), (2.3, ZXI), ...].

    Returns:
        The reduced stabilizer matrix and the set of independent Paulis.
    """
    # Construct the stabilizer matrix as rows
    stabilizer_rows = []
    for _, pauli in clique:
        cur_row = np.zeros(2 * N)
        for i in range(N):
            # Pauli Z's never appear in the Mermin operator
            if pauli[i] == "X":
                cur_row[i + N] = 1
            elif pauli[i] == "Y":
                cur_row[i] = 1
                cur_row[i + N] = 1
        stabilizer_rows.append(cur_row)

    # Transpose the matrix to put it in the proper Z+X form with shape (2*N, N)
    dependent_stabilizer = np.array(stabilizer_rows, dtype=np.uint8).T

    # Find the independent columns:
    # (1) Put the matrix in row Echelon form (mod 2)
    echelon_matrix = binary_gaussian_elimination(copy.copy(dependent_stabilizer))

    # (2) Keep only the independent columns
    stabilizer_matrix = []
    pauli_basis = []
    start_idx = 0
    for i, row in enumerate(echelon_matrix.T):
        matrix_row = np.asarray(row)
        if 1 in matrix_row[start_idx:]:
            stabilizer_matrix.append(dependent_stabilizer.T[i])
            pauli_basis.append(clique[i][1])
            last_1_idx = 0
            for j, val in enumerate(matrix_row):
                if val == 1:
                    last_1_idx = j
            start_idx = last_1_idx + 1

    return np.array(stabilizer_matrix).T, pauli_basis


def binary_gaussian_elimination(M: npt.NDArray[np.uint8]) -> npt.NDArray[np.uint8]:
    """Use binary Gaussian elimination to put the input matrix in row echelon form.

    The input matrix should be a binary matrix.

    Args:
        M: Input matrix that contains linearly depedent columns.

    Returns:
        A modified matrix in row echelon form.
    """
    num_rows, num_cols = M.shape
    for i in range(num_cols):
        if i >= num_rows:
            break
        # find the last non-zero element of column i, starting in row i
        max_i = i
        for k in range(i, num_rows):
            if M[k, i] == 1:
                max_i = k

        # Swap rows i and max_i
        M[[i, max_i], :] = M[[max_i, i], :]

        for u in range(i + 1, num_rows):
            # Add M[u, i] * row i to row u
            M[u, :] = (M[u, i] * M[i, :] + M[u, :]) % 2

    # Matrix M is now in row echelon form
    return M


def prepare_X_matrix(measurement_circuit: MeasurementCircuit) -> None:
    """Apply H's to a subset of qubits to ensure that the X matrix has full rank.

    Args:
        measurement_circuit: The current measurement circuit to act on.
    """
    # TODO: right now, this is naively trying all possibilities. Really, should apply
    # the polytime technique described in Aaronson https://arxiv.org/pdf/quant-ph/0406196.pdf
    N = measurement_circuit.num_qubits
    for bitstring in range(2**N):
        measurement_circuit_copy = copy.deepcopy(measurement_circuit)
        for i, bit in enumerate(f"{bitstring:b}".zfill(N)):
            if bit == "1":
                apply_H(measurement_circuit_copy, i)
        if (
            np.linalg.matrix_rank(measurement_circuit_copy.get_stabilizer()[N:]) == N
        ):  # done if full rank
            measurement_circuit.set_circuit(measurement_circuit_copy.get_circuit())
            measurement_circuit.set_stabilizer(measurement_circuit_copy.get_stabilizer())
            return


def row_reduce_X_matrix(measurement_circuit: MeasurementCircuit) -> None:
    """Use Gaussian elimination to reduce the Z matrix to the Identity matrix.

    Args:
        measurement_circuit: The current measurement circuit to act on.
    """
    transform_X_matrix_to_row_echelon_form(measurement_circuit)
    transform_X_matrix_to_reduced_row_echelon_form(measurement_circuit)


def transform_X_matrix_to_row_echelon_form(measurement_circuit: MeasurementCircuit) -> None:
    """Apply SWAPs and CNOTs until the X matrix is in row echelon form.

    Args:
        measurement_circuit: The current measurement circuit to act on.
    """
    N = measurement_circuit.num_qubits
    for j in range(N):
        if measurement_circuit.get_stabilizer()[j + N, j] == 0:
            i = j + 1
            while measurement_circuit.get_stabilizer()[i + N, j] == 0:
                i += 1
            apply_SWAP(measurement_circuit, i, j)

        for i in range(N + j + 1, 2 * N):
            if measurement_circuit.get_stabilizer()[i, j] == 1:
                apply_CNOT(measurement_circuit, j, i - N)


def transform_X_matrix_to_reduced_row_echelon_form(measurement_circuit: MeasurementCircuit) -> None:
    """Apply CNOTs to put the X matrix in reduced echelon form.

    The X stabilizer matrix of the input MeasurementCircuit should already be
    in row echelon form.

    Args:
        measurement_circuit: The current measurement circuit to act on.
    """
    N = measurement_circuit.num_qubits
    for j in range(N - 1, 0, -1):
        for i in range(N, N + j):
            if measurement_circuit.get_stabilizer()[i, j] == 1:
                apply_CNOT(measurement_circuit, j, i - N)


def patch_Z_matrix(measurement_circuit: MeasurementCircuit) -> None:
    """Apply S and CZ operations to clear the Z matrix.

    Args:
        measurement_circuit: The current measurement circuit to act on.
    """
    stabilizer_matrix, N = measurement_circuit.get_stabilizer(), measurement_circuit.num_qubits
    assert np.allclose(
        stabilizer_matrix[:N], stabilizer_matrix[:N].T
    ), f"Z-matrix,\n{stabilizer_matrix} is not symmetric"

    for i in range(N):
        for j in range(0, i):
            if stabilizer_matrix[i, j] == 1:
                apply_CZ(measurement_circuit, i, j)

        j = i
        if stabilizer_matrix[i, j] == 1:
            apply_S(measurement_circuit, i)


def change_X_to_Z_basis(measurement_circuit: MeasurementCircuit) -> None:
    """Apply Hadamards to swap the Z and X matrices.

    Args:
        measurement_circuit: The current measurement circuit to act on.
    """
    # change each qubit from X basis to Z basis via H
    N = measurement_circuit.num_qubits
    for j in range(N):
        apply_H(measurement_circuit, j)


def apply_H(measurement_circuit: MeasurementCircuit, i: int) -> None:
    """Apply a Hadamard on the specified qubit.

    Args:
        measurement_circuit: The current measurement circuit to act on.
        i: Index of the target qubit.
    """
    N = measurement_circuit.num_qubits
    qubits = measurement_circuit.qubits
    measurement_circuit.get_stabilizer()[[i, i + N]] = measurement_circuit.get_stabilizer()[
        [i + N, i]
    ]
    measurement_circuit.get_circuit().append(cirq.H(qubits[i]))


def apply_S(measurement_circuit: MeasurementCircuit, i: int) -> None:
    """Apply an S gate on the specified qubit.

    Args:
        measurement_circuit: The current measurement circuit to act on.
        i: Index of the target qubit.
    """
    qubits = measurement_circuit.qubits
    measurement_circuit.get_stabilizer()[i, i] = 0
    measurement_circuit.get_circuit().append(cirq.S(qubits[i]))


def apply_CZ(measurement_circuit: MeasurementCircuit, i: int, j: int) -> None:
    """Apply a CZ gate on the specified qubits.

    Args:
        measurement_circuit: The current measurement circuit to act on.
        i: Index of the control qubit.
        j: Index of the target qubit.
    """
    qubits = measurement_circuit.qubits
    measurement_circuit.get_stabilizer()[i, j] = 0
    measurement_circuit.get_stabilizer()[j, i] = 0
    measurement_circuit.get_circuit().append(cirq.CZ(qubits[i], qubits[j]))


def apply_CNOT(
    measurement_circuit: MeasurementCircuit, control_index: int, target_index: int
) -> None:
    """Apply a CNOT gate on the specified qubits.

    Args:
        measurement_circuit: The current measurement circuit to act on.
        control_index: Index of the control qubit.
        target_index: Index of the target qubit.
    """
    N = measurement_circuit.num_qubits
    qubits = measurement_circuit.qubits
    measurement_circuit.get_stabilizer()[control_index] = (
        measurement_circuit.get_stabilizer()[control_index]
        + measurement_circuit.get_stabilizer()[target_index]
    ) % 2
    measurement_circuit.get_stabilizer()[target_index + N] = (
        measurement_circuit.get_stabilizer()[control_index + N]
        + measurement_circuit.get_stabilizer()[target_index + N]
    ) % 2
    measurement_circuit.get_circuit().append(cirq.CNOT(qubits[control_index], qubits[target_index]))


def apply_SWAP(measurement_circuit: MeasurementCircuit, i: int, j: int) -> None:
    """Apply a SWAP gate on the specified qubits.

    Args:
        measurement_circuit: The current measurement circuit to act on.
        i: Index of the control qubit.
        j: Index of the target qubit.
    """
    N = measurement_circuit.num_qubits
    qubits = measurement_circuit.qubits
    measurement_circuit.get_stabilizer()[[i, j]] = measurement_circuit.get_stabilizer()[[j, i]]
    measurement_circuit.get_stabilizer()[[i + N, j + N]] = measurement_circuit.get_stabilizer()[
        [j + N, i + N]
    ]
    measurement_circuit.get_circuit().append(cirq.SWAP(qubits[i], qubits[j]))



from __future__ import annotations

import cirq

import supermarq


def compute_communication(circuit: cirq.Circuit) -> float:
    """Compute the *communication* feature of the input circuit.

    This function acts a wrapper which first converts the input `cirq.Circuit`
    into a `qiskit.QuantumCircuit` before calculating the feature value.

    Args:
        circuit: A quantum circuit.

    Returns:
        The value of the communication feature for this circuit.
    """
    return supermarq.converters.compute_communication_with_qiskit(
        supermarq.converters.cirq_to_qiskit(circuit)
    )


def compute_liveness(circuit: cirq.Circuit) -> float:
    """Compute the *liveness* feature of the input circuit.

    This function acts a wrapper which first converts the input `cirq.Circuit`
    into a `qiskit.QuantumCircuit` before calculating the feature value.

    Args:
        circuit: A quantum circuit.

    Returns:
        The value of the liveness feature for this circuit.
    """
    return supermarq.converters.compute_liveness_with_qiskit(
        supermarq.converters.cirq_to_qiskit(circuit)
    )


def compute_parallelism(circuit: cirq.Circuit) -> float:
    """Compute the *parallelism* feature of the input circuit.

    This function acts a wrapper which first converts the input `cirq.Circuit`
    into a `qiskit.QuantumCircuit` before calculating the feature value.

    Args:
        circuit: A quantum circuit.

    Returns:
        The value of the parallelism feature for this circuit.
    """
    return supermarq.converters.compute_parallelism_with_qiskit(
        supermarq.converters.cirq_to_qiskit(circuit)
    )


def compute_measurement(circuit: cirq.Circuit) -> float:
    """Compute the *measurement* feature of the input circuit.

    This function acts a wrapper which first converts the input `cirq.Circuit`
    into a `qiskit.QuantumCircuit` before calculating the feature value.

    Args:
        circuit: A quantum circuit.

    Returns:
        The value of the measurement feature for this circuit.
    """
    return supermarq.converters.compute_measurement_with_qiskit(
        supermarq.converters.cirq_to_qiskit(circuit)
    )


def compute_entanglement(circuit: cirq.Circuit) -> float:
    """Compute the *entanglement* feature of the input circuit.

    This function acts a wrapper which first converts the input `cirq.Circuit`
    into a `qiskit.QuantumCircuit` before calculating the feature value.

    Args:
        circuit: A quantum circuit.

    Returns:
        The value of the entanglement feature for this circuit.
    """
    return supermarq.converters.compute_entanglement_with_qiskit(
        supermarq.converters.cirq_to_qiskit(circuit)
    )


def compute_depth(circuit: cirq.Circuit) -> float:
    """Compute the *depth* feature of the input circuit.

    This function acts a wrapper which first converts the input `cirq.Circuit`
    into a `qiskit.QuantumCircuit` before calculating the feature value.

    Args:
        circuit: A quantum circuit.

    Returns:
        The value of the depth feature for this circuit.
    """
    return supermarq.converters.compute_depth_with_qiskit(
        supermarq.converters.cirq_to_qiskit(circuit)
    )



# pylint: disable=missing-function-docstring,missing-class-docstring
from __future__ import annotations

import cirq
import qiskit

import supermarq


def test_cirq_to_qiskit() -> None:
    cirq_circuit = cirq.Circuit(
        cirq.H(cirq.LineQubit(0)), cirq.CX(cirq.LineQubit(0), cirq.LineQubit(1))
    )
    qiskit_circuit = qiskit.QuantumCircuit(2)
    qiskit_circuit.h(0)
    qiskit_circuit.cx(0, 1)
    assert supermarq.converters.cirq_to_qiskit(cirq_circuit) == qiskit_circuit


def test_compute_parallelism_with_qiskit() -> None:
    qiskit_circuit = qiskit.QuantumCircuit(2)
    assert supermarq.converters.compute_parallelism_with_qiskit(qiskit_circuit) == 0



# Copyright 2021 The Cirq Developers
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base experiment class and tools used across all experiments."""
from __future__ import annotations

import functools
import numbers
import pathlib
import uuid
import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Generic, TypeVar

import cirq
import cirq_superstaq as css
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tqdm.auto import tqdm

import supermarq

if TYPE_CHECKING:
    from typing_extensions import Self


def qcvv_resolver(cirq_type: str) -> type[Any] | None:
    """Resolves string's referencing classes in the QCVV library. Used by `cirq.read_json()`
    to deserialize.

    Args:
        cirq_type: The type being resolved

    Returns:
        The corresponding type object (if found) else None

    Raises:
        ValueError: If the provided type is not resolvable
    """
    prefix = "supermarq.qcvv."
    if cirq_type.startswith(prefix):
        name = cirq_type[len(prefix) :]
        if name in supermarq.qcvv.__all__:
            return getattr(supermarq.qcvv, name, None)
    return None


@dataclass
class Sample:
    """A sample circuit to use along with any data about the circuit
    that is needed for analysis
    """

    circuit_realization: int
    """Indicates which realization of the random circuit this sample is. There will be D samples
    with matching circuit realization value, one for each cycle depth being measured. This index is
    useful for grouping results during analysis.
    """
    circuit: cirq.Circuit
    """The raw (i.e. pre-compiled) sample circuit."""
    data: dict[str, Any]
    """The corresponding data about the circuit that is needed when analyzing results
    (e.g. cycle depth)."""

    uuid: uuid.UUID = field(default_factory=uuid.uuid1)
    """The unique ID of the sample."""

    def __hash__(self) -> int:
        return hash(
            (
                self.circuit_realization,
                self.uuid,
                self.circuit.freeze(),
                tuple(sorted(self.data.items())),
            )
        )

    def _json_dict_(self) -> dict[str, Any]:
        """Converts the sample to a json-able dictionary that can be used to recreate the
        sample object.

        Returns:
            Json-able dictionary of the sample data.
        """
        return {
            "circuit": self.circuit,
            "data": self.data,
            "circuit_realization": self.circuit_realization,
            "sample_uuid": str(self.uuid),
        }

    @classmethod
    def _from_json_dict_(
        cls,
        circuit: cirq.Circuit,
        circuit_realization: int,
        data: dict[str, Any],
        sample_uuid: str,
        **_: Any,
    ) -> Self:
        """Creates a sample from a dictionary of the data.

        Args:
            dictionary: Dict containing the sample data.

        Returns:
            The deserialized Sample object.
        """
        return cls(
            circuit=circuit,
            circuit_realization=circuit_realization,
            data=data,
            uuid=uuid.UUID(sample_uuid),
        )

    @classmethod
    def _json_namespace_(cls) -> str:
        return "supermarq.qcvv"


@dataclass
class QCVVResults(ABC):
    """A dataclass for storing the data and analyze results of the experiment. Requires
    subclassing for each new experiment type."""

    target: str
    """The target device that was used."""

    experiment: QCVVExperiment[QCVVResults]
    """Reference to the underlying experiment that generated these results experiment."""

    job: css.Job | None = None
    """The associated Superstaq job (if applicable)."""

    data: pd.DataFrame | None = None
    """The raw data generated."""

    @property
    def data_ready(self) -> bool:
        """Whether the experimental data is ready to analyse.

        Raises:
            RuntimeError: If their is no stored data and no Superstaq job to use to collect the
                results.
        """
        if self.data is not None:
            return True
        if self.job is None:
            raise RuntimeError(
                "No data available and no Superstaq job to use to collect data. Please manually "
                "add results data in order to perform analysis"
            )
        job_status = self.job.status()
        if job_status == "Done":
            self.data = self._collect_device_counts()
            return True
        return False

    @property
    def samples(self) -> Sequence[Sample]:
        """Returns:
        The number of samples used."""
        return self.experiment.samples

    @property
    def num_qubits(self) -> int:
        """Returns:
        The number of qubits in the experiment."""
        return self.experiment.num_qubits

    @property
    def num_circuits(self) -> int:
        """Returns:
        The number of circuits in the experiment."""
        return self.experiment.num_circuits

    def analyze(
        self,
        plot_results: bool = True,
        print_results: bool = True,
        plot_filename: str | None = None,
    ) -> None:
        """Perform the experiment analysis and store the results in the `results` attribute.

        Args:
            plot_results: Whether to generate plots of the results. Defaults to True.
            print_results: Whether to print the final results. Defaults to True.
            plot_filename: Optional argument providing a filename to save the plots to. Ignored if
                `plot_results=False` Defaults to None, indicating not to save the plot.
        """
        if not self.data_ready:
            warnings.warn(
                "Experiment data is not yet ready to analyse. This is likely because "
                "the Superstaq job has not yet been completed. Either wait and try again "
                "later, or interrogate the `.job` attribute."
            )
            return

        self._analyze()

        if plot_results:
            self.plot_results(filename=plot_filename)

        if print_results:
            self.print_results()

    @abstractmethod
    def _analyze(self) -> None:
        """A method that analyses the `data` attribute and stores the final experimental results."""

    @abstractmethod
    def plot_results(self, filename: str | None = None) -> plt.Figure:
        """Plot the results of the experiment

        Args:
            filename: Optional argument providing a filename to save the plots to. Defaults to None,
                indicating not to save the plot.
        Returns:
            A single matplotlib figure containing the relevant plots of the results data.
        """

    @abstractmethod
    def print_results(self) -> None:
        """Prints the key results data."""

    def _collect_device_counts(self) -> pd.DataFrame:
        """Process the counts returned by the server and process into a results dataframe.

        Returns:
            The results dataframe.
        """
        if self.job is None:
            raise ValueError(
                "No Superstaq job associated with these results. Cannot collect device counts."
            )
        records = []
        device_counts = self.job.counts()
        for counts, sample in zip(device_counts, self.samples):

            total = sum(counts.values())
            probabilities = {
                format(idx, f"0{self.num_qubits}b"): 0.0 for idx in range(2**self.num_qubits)
            }
            for key, count in counts.items():
                probabilities[key] = count / total
            records.append({**sample.data, **probabilities})

        return pd.DataFrame(records)

    @property
    def _not_analyzed(self) -> RuntimeError:
        return RuntimeError("Value has not yet been estimated. Please run `.analyze()` method.")


ResultsT = TypeVar("ResultsT", bound=QCVVResults, covariant=True)
# Generic results type for base experiments.


class QCVVExperiment(ABC, Generic[ResultsT]):
    """Base class for gate benchmarking experiments.

    The interface for implementing these experiments is as follows:

    #. First instantiate the desired experiment object

        .. code::

            experiment = ExampleExperiment(<<args/kwargs>>)

    #. Prepare the circuits and run the experiment on the desired target. This can either be a
       custom simulator or a real device name. For example:

        .. code::

            noise_model = cirq.depolarize(p=0.01, n_qubits=1)
            sim = cirq.DensityMatrixSimulator(noise=noise_model)

            results = experiment.run_with_simulator(simulator=sim, <<args/kwargs>>)

    #. Then we analyse the results. If the target was a local simulator this will be available as
       soon as the :code:`run_with_simulator()` method has finished executing. On the other hand
       if a real device was accessed via Superstaq then it may take time for the data to be
       available from the server. The :code:`results.data_ready` attribute will return
       :code:`True` when all data has been collected and is ready to be analyzed.

       .. code::

            if results.data_ready():
                results.analyze(<<args>>)

    When implementing a new experiment, 4 methods need to be implemented:

    #. :meth:`experiment._build_circuits()`: Given a number of circuits and an iterable of the
        different numbers of layers to use, return a list of :class:`Sample` objects that need to
        be sampled during the experiment.

    #. :meth:`results._analyse_results()`: Analyse the experimental data and store the final
        results, for example some fidelities.

    #. :meth:`results.plot_results()`:  Produce any relevant plots that are useful for understanding
        the results of the experiment.

    #. :meth:`results.print_results()`: Prints the results to the console.
    """

    def __init__(
        self,
        qubits: int | Sequence[cirq.Qid],
        num_circuits: int,
        cycle_depths: Iterable[int],
        *,
        random_seed: int | np.random.Generator | None = None,
        results_cls: type[ResultsT],
        _samples: Sequence[Sample] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initializes a benchmarking experiment.

        Args:
            qubits: The qubits used during the experiment. If an integer, this number of line qubits
                will be used. Most subclasses will determine this from their other inputs.
            num_circuits: The number of circuits to sample.
            cycle_depths: A sequence of depths to sample.
            random_seed: An optional seed to use for randomization.
            results_cls: The results class to use for the experiment.
            _samples: Optional list of samples to construct the experiment from
            kwargs: Additional kwargs passed to the Superstaq service object.
        """
        self.qubits: Sequence[cirq.Qid]
        if isinstance(qubits, Sequence):
            self.qubits = list(qubits)
        else:
            self.qubits = cirq.LineQubit.range(qubits)

        """The qubits used in the experiment."""

        self.num_circuits = num_circuits
        """The number of circuits to build for each cycle depth."""

        self.cycle_depths = list(cycle_depths)
        """The different cycle depths to test at."""

        self._service_kwargs = kwargs
        """Arguments to pass to the Superstaq service for submitting jobs."""

        self._rng = np.random.default_rng(random_seed)

        self._results_cls: type[ResultsT] = results_cls

        if not _samples:
            self.samples = self._prepare_experiment()
        else:
            self.samples = _samples
        """Create all the samples needed for the experiment."""

    def __getitem__(self, key: str | int | uuid.UUID) -> Sample:
        """Gets a sample from the experiment using a key which is either an int (representing the
        index of the circuit) or a str/uuid.UUID (representing the sample's UUID).

        Args:
            key: The key of the sample to find.

        Raises:
            TypeError: If the key is not an int, str or uuid.UUID
            KeyError: If matching Sample can be found
            RuntimeError: If multiple samples are found with the same key.

        Returns:
            The sample corresponding to the key.
        """
        if isinstance(key, numbers.Integral):
            return self.samples[key]
        elif isinstance(key, str):
            key = uuid.UUID(key)
        elif not isinstance(key, uuid.UUID):
            raise TypeError(f"Key must be int, str or uuid.UUID, not {type(key)}")

        matching_samples = [s for s in self.samples if s.uuid == key]
        if len(matching_samples) == 1:
            return matching_samples[0]
        elif len(matching_samples) == 0:
            raise KeyError(f"No sample found with UUID {key}")
        else:
            raise RuntimeError(
                "Multiple samples found with matching key. Something has gone wrong."
            )

    def __iter__(self) -> Iterator[Sample]:
        return iter(self.samples)

    ##############
    # Properties #
    ##############
    @functools.cached_property
    def _superstaq_service(self) -> css.Service:
        """A Superstaq service to use for compilation and circuit submission."""
        return css.Service(**self._service_kwargs)

    @property
    def num_qubits(self) -> int:
        """Returns:
        The number of qubits used in the experiment
        """
        return len(self.qubits)

    ###################
    # Private Methods #
    ###################
    @abstractmethod
    def _build_circuits(
        self,
        num_circuits: int,
        cycle_depths: Iterable[int],
    ) -> Sequence[Sample]:
        """Build a list of circuits required for the experiment. These circuits are stored in
        :class:`Sample` objects along with any additional data that is needed during the analysis.

        Args:
            num_circuits: Number of circuits to generate.
            cycle_depths: An iterable of the different cycle depths to use during the experiment.

        Returns:
           The list of experiment samples.
        """

    @staticmethod
    def canonicalize_bitstring(key: int | str, num_qubits: int) -> str:
        """Checks that the provided key represents a bit string for the given number of qubits.
        If the key is provided as an integer then this is reformatted as a bitstring.

        Args:
            key: The integer or string which represents a bitstring.
            num_qubits: The number of bits that the bitstring needs to have

        Raises:
            ValueError: If the key is integer and negative
            ValueError: If the key is integer but to large for the given number of qubits.
            ValueError: If the key is a string but the wrong length.
            ValueError: If the key is a string but contains characters that are not 0 or 1.
            TypeError: If the key value is not a string or integral.

        Returns:
            The canonicalized representation of the bitstring.
        """
        if isinstance(key, numbers.Integral):
            if key < 0:
                raise ValueError(f"The key must be positive. Instead got {key}.")
            if key >= 2**num_qubits:
                raise ValueError(
                    f"The key is too large to be encoded with {num_qubits} qubits. Got {key} "
                    f"but expected less than {2**num_qubits}."
                )
            return format(key, f"0{num_qubits}b")

        if isinstance(key, str):
            if len(key) != num_qubits:
                raise ValueError(
                    f"The key contains the wrong number of bits. Got {len(key)} entries "
                    f"but expected {num_qubits} bits."
                )
            if any(b not in ["0", "1"] for b in key):
                raise ValueError(f"All entries in the bitstring must be 0 or 1. Got {key}.")
            return key

        raise TypeError("Key must either be `numbers.Integral` or `str`.")

    @staticmethod
    def canonicalize_probabilities(
        results: Mapping[str, float] | Mapping[int, float],
        num_qubits: int,
    ) -> dict[str, float]:
        """Reformats a dictionary of probabilities/counts so that all keys are bitstrings and that
        there are no missing values. Also renormalizes so that the resulting probabilities sum to 1
        and sorts the dictionary by bitstring.

        Args:
            results: The unformatted probabilities or counts
            num_qubits: The number of qubits, used to determine the bitstring length.

        Raises:
            ValueError: If any counts or probabilities are negative.
            ValueError: If there are no non-zero counts.

        Returns:
            The formatted dictionary of probabilities.
        """
        if not results:
            return {}

        if any(c < 0 for c in results.values()):
            raise ValueError("Probabilities/counts must be positive.")
        if sum(results.values()) == 0:
            raise ValueError("No non-zero counts.")
        probabilities = {
            QCVVExperiment.canonicalize_bitstring(key, num_qubits): count / sum(results.values())
            for key, count in results.items()
        }
        # Add zero values for any missing bitstrings
        for k in range(2**num_qubits):
            if (bitstring := format(k, f"0{num_qubits}b")) not in probabilities:
                probabilities[bitstring] = 0.0
        # Sort by bitstrings
        probabilities = dict(sorted(probabilities.items()))

        return probabilities

    @staticmethod
    def _interleave_op(
        circuit: cirq.Circuit, operation: cirq.Operation, include_final: bool = False
    ) -> cirq.Circuit:
        """Interleave a given operation into a circuit.

        Args:
            circuit: The original circuit.
            operation: The operation to interleave.
            include_final: If True then the interleaving gate is also appended to
                the end of the circuit.

        Returns:
            A copy of the original circuit with the provided gate interleaved.
        """
        operation = operation.with_tags("no_compile")
        interleaved_circuit = circuit.copy()
        interleaved_circuit.batch_insert(
            [(k, operation) for k in range(len(circuit) - int(not include_final), 0, -1)]
        )
        return interleaved_circuit

    def _map_records_to_samples(
        self, records: Mapping[uuid.UUID | int, Mapping[str, float] | Mapping[int, float]]
    ) -> dict[Sample, Mapping[str, float] | Mapping[int, float]]:
        """Creates a mapping between experiment samples and the provided results records. Records
        with unrecognized sample keys (which should be either an integer index or a UUID) are
        ignored.

        Args:
            records: A mapping of sample keys (either an integer index or a UUID for the sample) to
                the corresponding bitcount/probability results.

        Returns:
            A mapping between experiment samples and the provided results records
        """
        records_not_mapped = dict(records)

        record_mapping: dict[Sample, Mapping[str, float] | Mapping[int, float]] = {}
        for key, record in records.items():
            try:
                sample = self[key]
            except (KeyError, IndexError):  # Ignore any keys that cant be attached to samples
                continue

            if sample in record_mapping:
                raise KeyError(f"Duplicate records found for sample with uuid: {str(sample.uuid)}.")
            record_mapping[sample] = record
            records_not_mapped.pop(key)

        missing_samples = [s for s in self if s not in record_mapping]
        if missing_samples:
            warnings.warn(
                "The following samples are missing records: "
                f"{', '.join(str(s.uuid) for s in missing_samples)}. These will not be included in "
                "the results.",
                stacklevel=2,
            )
        if records_not_mapped:
            warnings.warn(
                f"Unable to find matching sample for {len(records_not_mapped)} record(s).",
                stacklevel=2,
            )

        return record_mapping

    @abstractmethod
    def _json_dict_(self) -> dict[str, Any]:
        """Converts the experiment to a json-able dictionary that can be used to recreate the
        experiment object. Note that the state of the random number generator is not stored.

        .. note:: Must be re-implemented in any subclasses to ensure all important data is stored.

        Returns:
            Json-able dictionary of the experiment data.
        """
        return {
            "cycle_depths": self.cycle_depths,
            "num_circuits": self.num_circuits,
            "qubits": self.qubits,
            "samples": self.samples,
            **self._service_kwargs,
        }

    @classmethod
    @abstractmethod
    def _from_json_dict_(cls, *args: Any, **kwargs: Any) -> Self:
        """Creates a experiment from an expanded dictionary of the data.

        Returns:
            The deserialized experiment object.
        """

    @classmethod
    def _json_namespace_(cls) -> str:
        return "supermarq.qcvv"

    def to_file(self, filename: str | pathlib.Path) -> None:
        """Save the experiment to a json file.

        Args:
            filename: Filename to save to.
        """
        with open(filename, "w") as file_stream:
            cirq.to_json(self, file_stream)

    @classmethod
    def from_file(cls, filename: str | pathlib.Path) -> Self:
        """Load the experiment from a json file.

        Args:
            filename: Filename to load from.

        Returns:
            The loaded experiment.
        """
        with open(filename, "r") as file_stream:
            experiment = cirq.read_json(
                file_stream,
                resolvers=[*css.SUPERSTAQ_RESOLVERS, *cirq.DEFAULT_RESOLVERS, qcvv_resolver],
            )
        return experiment

    def _prepare_experiment(
        self,
    ) -> Sequence[Sample]:
        """Prepares the circuits needed for the experiment

        Args:
            num_circuits: Number of circuits to run.
            cycle_depths: An iterable of the different layer depths to use during the experiment.
            overwrite: Whether to force an experiment run even if there is existing data that would
                be over written in the process. Defaults to False.

        Raises:
            RuntimeError: If the experiment has already been run once and the `overwrite` argument
                is not True
            ValueError: If any of the cycle depths provided negative or zero.

        Returns:
            A sequence of samples for the experiment.
        """

        if any(depth <= 0 for depth in self.cycle_depths):
            raise ValueError("The `cycle_depths` iterator can only include positive values.")

        samples = self._build_circuits(self.num_circuits, self.cycle_depths)
        self._validate_circuits(samples)
        return samples

    def _validate_circuits(self, samples: Sequence[Sample]) -> None:
        """Checks that all circuits contain a terminal measurement of all qubits.

        Args:
            samples: The sequence of samples to check.
        """
        for sample in samples:
            if not sample.circuit.has_measurements():
                raise ValueError("QCVV experiment circuits must contain measurements.")
            if not sample.circuit.are_all_measurements_terminal():
                raise ValueError("QCVV experiment circuits can only contain terminal measurements.")
            if not sorted(sample.circuit[-1].qubits) == sorted(self.qubits):
                raise ValueError(
                    "The terminal measurement in QCVV experiment circuits must measure all qubits."
                )

    ###################
    # Public Methods  #
    ###################
    def run_on_device(
        self,
        target: str,
        repetitions: int = 10_000,
        method: str | None = None,
        **target_options: Any,
    ) -> ResultsT:
        """Submit the circuit samples to the desired target device and store the resulting
        probabilities.

        The set of circuits is partitioned as necessary to not exceed the maximum circuits that can
        be submitted to the given target device. The function then waits for the jobs to complete
        before saving the resulting probability distributions.

        Args:
            target: The name of a Superstaq target.
            repetitions: The number of shots to sample. Defaults to 10,000.
            method: Optional method to use on the Superstaq device. Defaults to None corresponding
                to normal running.
            target_options: Optional configuration dictionary passed when submitting the job.

        Returns:
            The experiment results object.
        """

        experiment_job = self._superstaq_service.create_job(
            [sample.circuit for sample in self.samples],
            target=target,
            method=method,
            repetitions=repetitions,
            **target_options,
        )

        return self._results_cls(
            target=target,
            experiment=self,
            job=experiment_job,
        )

    def run_with_simulator(
        self,
        simulator: cirq.Sampler | None = None,
        repetitions: int = 10_000,
    ) -> ResultsT:
        """Use the local simulator to sample the circuits and store the resulting probabilities.

        Args:
            simulator: A local :class:`~cirq.Sampler` to use. If None then the default
                :class:`cirq.Simulator` simulator is used. Defaults to None.
            repetitions: The number of shots to sample. Defaults to 10,000.

        Returns:
            The experiment results object.
        """
        if simulator is None:
            simulator = cirq.Simulator(seed=self._rng)

        records = []
        for sample in tqdm(self.samples, desc="Simulating circuits"):
            result = simulator.run(sample.circuit, repetitions=repetitions)
            hist = result.histogram(key=cirq.measurement_key_name(sample.circuit))
            probabilities = self.canonicalize_probabilities(
                {key: count / sum(hist.values()) for key, count in hist.items()}, self.num_qubits
            )
            records.append(
                {"circuit_realization": sample.circuit_realization, **sample.data, **probabilities}
            )

        return self._results_cls(
            target="local_simulator",
            experiment=self,
            data=pd.DataFrame(records),
        )

    def run_with_callable(
        self,
        circuit_eval_func: Callable[[cirq.Circuit], Mapping[str, float] | Mapping[int, float]],
        **kwargs: Any,
    ) -> ResultsT:
        """Evaluates the probabilities for each circuit using a user provided callable function.
        This function should take a circuit as input and return a dictionary of probabilities for
        each bitstring (including states with zero probability).

        Args:
            circuit_eval_func: The custom function to use when evaluating circuit probabilities.
            kwargs: Additional arguments to pass to the custom function.

        Returns:
            The experiment results object.
        """
        records = []
        for sample in tqdm(self.samples, desc="Running circuits"):
            raw_probability = circuit_eval_func(sample.circuit, **kwargs)
            probability = self.canonicalize_probabilities(raw_probability, self.num_qubits)
            records.append({**sample.data, **probability})

        return self._results_cls(
            target="callable",
            experiment=self,
            data=pd.DataFrame(records),
        )

    def results_from_records(
        self,
        records: Mapping[uuid.UUID | int, Mapping[str, float] | Mapping[int, float]],
    ) -> ResultsT:
        """Creates a results object from records of the counts/probabilities for each sample
        circuit. This function is aimed at users who would like to use the QCVV framework to
        generate sample circuits and analyse the results but need to run these circuits without
        submitting a job to Superstaq.

        Args:
            records: A dictionary of the counts/probabilities for each sample, keyed by either the
                sample UUID or the index of the sample in the experiment. The counts/probabilities
                for each sample should be provided as a
                dictionary of integers or floats (respectively) keyed by either the bitstring or
                the integer value of that bitstring. Note that the distinction between counts and
                probabilities is inferred from the type (int vs float respectively). Please do not
                use float type for counts.

        Returns:
            The experiment results object.
        """
        sample_mapping = self._map_records_to_samples(records)

        results_data = []
        for sample, results in sample_mapping.items():
            probabilities = self.canonicalize_probabilities(
                results,
                self.num_qubits,
            )

            # Add to results data
            results_data.append({**sample.data, **probabilities})

        return self._results_cls(
            target="records",
            experiment=self,
            data=pd.DataFrame(results_data),
        )
